from typing import Dict, Any, Tuple, List
import boto3
import math
from tqdm import tqdm

from blue_options import string
from blueflow.workflow.generic import Workflow
from blueflow.logger import logger
from blueflow.workflow.runners.generic import GenericRunner
from blueflow.aws_batch.submission import submit, SubmissionType


class AWSBatchRunner(GenericRunner):
    def __init__(self, **kw_args):
        super().__init__(**kw_args)
        self.type_name: str = "aws_batch"

    def monitor_function(
        self,
        workflow: Workflow,
        hot_node: str,
    ) -> Workflow:
        workflow = super().monitor_function(workflow, hot_node)

        client = boto3.client("batch")

        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch/client/describe_jobs.html
        page_size = 100
        for index in tqdm(range(0, len(workflow.G.nodes), page_size)):
            nodes = list(workflow.G.nodes)[index : index + page_size]

            jobs = [
                workflow.G.nodes[node].get("job_id").replace('"', "") for node in nodes
            ]

            response = client.describe_jobs(jobs=jobs)

            status: Dict[str, str] = {}
            for item in response["jobs"]:
                status[item["jobId"]] = item["status"]

            for node, job_id in zip(nodes, jobs):
                workflow.G.nodes[node]["status"] = status[job_id]

        return workflow

    def submit(
        self,
        workflow: Workflow,
        dryrun: bool = True,
        max_dependency: int = 20,
    ) -> bool:
        list_of_nodes = list(workflow.G.nodes.keys())
        for node in list_of_nodes:
            dependency_list = list(workflow.G.successors(node))
            if len(dependency_list) <= max_dependency:
                continue

            proxy_count = math.ceil(len(dependency_list) / (max_dependency - 1)) - 1
            suffix = string.random(length=3, alphabet="0123456789")
            logger.info(
                "@ {}: {} dependencies > {} - adding {} proxy(s) @ {}.".format(
                    node,
                    len(dependency_list),
                    max_dependency,
                    proxy_count,
                    suffix,
                )
            )

            proxy_list = [
                "{}-{}-{:03d}".format(node, suffix, index + 1)
                for index in range(proxy_count)
            ]
            for proxy in proxy_list:
                workflow.G.add_node(proxy)
                workflow.G.nodes[proxy]["command_line"] = " ".join(
                    [
                        "workflow monitor",
                        f"node={proxy}",
                        self.job_name,
                    ]
                )

            previous_proxy = node
            for index, node_ in enumerate(dependency_list):
                proxy_index = int(index / (max_dependency - 1))
                if proxy_index < 1:
                    continue

                proxy = proxy_list[proxy_index - 1]

                workflow.G.remove_edge(node, node_)

                if proxy != previous_proxy:
                    workflow.G.add_edge(previous_proxy, proxy)
                    previous_proxy = proxy

                workflow.G.add_edge(proxy, node_)

                logger.info(
                    "{}->{} -> {}->{}->{}".format(
                        node,
                        node_,
                        node,
                        proxy,
                        node_,
                    )
                )

        return workflow.save() and super().submit(workflow, dryrun)

    def submit_command(
        self,
        command_line: str,
        job_name: str,
        dependencies: List[str],
        verbose: bool = False,
        type: str = "cpu",
    ) -> Tuple[bool, Any]:
        assert super().submit_command(
            command_line,
            job_name,
            dependencies,
            verbose,
            type,
        )[0]

        # ignore type - all jobs are cpu.
        return submit(
            f"- {command_line}",
            job_name,
            SubmissionType.EVAL,
            dependency_job_id_list=dependencies,
            verbose=verbose,
        )



import sys
import argparse

from blueness import module
from blueness.argparse.generic import sys_exit
from blue_objects import file

from blueflow import NAME
from blueflow.aws_batch.submission import (
    submit,
    SubmissionType,
)
from blueflow.logger import logger

NAME = module.name(__file__, NAME)


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="cat_log|get_log_stream_name|show_count|submit",
)
parser.add_argument(
    "--command_line",
    type=str,
    default="",
)
parser.add_argument(
    "--job_name",
    type=str,
    default="",
)
parser.add_argument(
    "--type",
    type=str,
    default="source",
    help="eval|source|submit",
)
parser.add_argument(
    "--filename",
    type=str,
)
args = parser.parse_args()

success = False
if args.task == "cat_log":
    success, content = file.load_json(args.filename, ignore_error=True)

    if success:
        count = 0
        for line in content.get("events", []):
            print(line.get("message", ""))
            count = count + 1

        print(f"📜 {count} line(s).")
elif args.task == "get_log_stream_name":
    success, metadata = file.load_json(args.filename)
    if success:
        print(
            metadata["jobs"][0].get("container", {}).get("logStreamName", "")
            if metadata["jobs"]
            else ""
        )
elif args.task == "show_count":
    success = True
    input_string = sys.stdin.read().strip()

    if not input_string.isdigit():
        print(input_string)
    else:
        input_int = int(input_string)
        if input_int:
            print("{} {}".format(input_int, input_int * "🌀"))
elif args.task == "submit":
    success, _ = submit(
        command_line=args.command_line,
        job_name=args.job_name,
        type=SubmissionType[args.type.upper()],
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Tuple, Any, List
from enum import Enum
import boto3

from blue_objects.env import ABCLI_AWS_REGION

from blueflow.logger import logger


class SubmissionType(Enum):
    EVAL = "eval"
    SOURCE = "source"
    SUBMIT = "submit"

    @property
    def runner(self):
        return "abcli_eval" if self == SubmissionType.EVAL else "abcli_scripts source"


def submit(
    command_line: str,
    job_name: str,
    type: SubmissionType,
    dependency_job_id_list: List[str] = [],
    verbose: bool = True,
) -> Tuple[bool, Any]:
    # https://unix.stackexchange.com/questions/243571/how-to-run-source-with-docker-exec/243580#243580
    command = [
        "bash",
        "-c",
        "source /root/git/awesome-bash-cli/abcli/.abcli/abcli.sh mono,install,aws_batch {} {}".format(
            type.runner,
            command_line,
        ),
    ]
    logger.info(
        "{} -> {}: {}{}".format(
            type.name,
            job_name,
            " 🔹 ".join(command),
            (
                " # {} dependenci(es): {}".format(
                    len(dependency_job_id_list),
                    ", ".join(dependency_job_id_list),
                )
                if dependency_job_id_list
                else ""
            ),
        )
    )

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html
    client = boto3.client("batch")

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch/client/submit_job.html
    response = client.submit_job(
        jobName=job_name,
        jobQueue="abcli-v3",
        jobDefinition="abcli-custom-v1",
        containerOverrides={
            "command": command,
        },
        dependsOn=[
            {
                "jobId": job_id,
                "type": "SEQUENTIAL",
            }
            for job_id in dependency_job_id_list
        ],
    )
    if verbose:
        logger.info(f"response: {response}")

    job_id = response["jobId"]
    if verbose:
        logger.info(f"job_id: {job_id}")
        logger.info(
            "🔗 https://{}.console.aws.amazon.com/batch/home?region={}#jobs/detail/{}".format(
                ABCLI_AWS_REGION,
                ABCLI_AWS_REGION,
                job_id,
            )
        )

    # for compatibility with other runners
    response["job_id"] = job_id

    return True, response



import os

from blue_options.env import load_config, load_env

load_env(__name__)
load_config(__name__)

ABCLI_AWS_BATCH_JOB_STATUS_LIST = os.getenv("ABCLI_AWS_BATCH_JOB_STATUS_LIST", "")
ABCLI_AWS_BATCH_JOB_STATUS_LIST_WATCH = os.getenv(
    "ABCLI_AWS_BATCH_JOB_STATUS_LIST_WATCH", ""
)

NBS_DEFAULT_WORKFLOW_PATTERN = os.getenv("NBS_DEFAULT_WORKFLOW_PATTERN", "")



from functools import reduce

from blue_options import string
from blue_objects.env import ABCLI_PUBLIC_PREFIX

from blueflow.workflow.patterns import list_of_patterns
from blueflow.workflow.runners import list_of_runners


items = (
    ["📜"]
    + [
        "[`{}`](./patterns/{}.dot)".format(
            pattern,
            pattern,
        )
        for pattern in list_of_patterns()
    ]
    + reduce(
        lambda x, y: x + y,
        [
            (
                [f"[{runner_type}](./runners/{runner_type}.py)"]
                + [
                    f"[![image]({url})]({url})"
                    for url in [
                        "{}/{}-{}/workflow.gif?raw=true&random={}".format(
                            ABCLI_PUBLIC_PREFIX,
                            runner_type,
                            pattern,
                            string.random(),
                        )
                        for pattern in list_of_patterns()
                    ]
                ]
            )
            for runner_type in list_of_runners()
        ],
        [],
    )
)



NAME = "blueflow"

ICON = "📜"

DESCRIPTION = f"{ICON} workflow management on aws batch and more."

VERSION = "4.769.1"

REPO_NAME = "notebooks-and-scripts"

MARQUEE = "https://github.com/kamangir/assets/raw/main/nbs/3x4.jpg?raw=true"


def fullname() -> str:
    return f"{NAME}-{VERSION}"



import os

from blue_objects import file, README
from blue_objects.env import ABCLI_PUBLIC_PREFIX

from blueflow import NAME, VERSION, ICON, REPO_NAME
from blueflow.workflow.README import items as workflow_items


features = {
    "workflow": {
        "description": "an abstraction to run mixed-type (cpu/gpu/...) [DAG](https://networkx.org/documentation/stable/reference/classes/digraph.html)s of commands on [aws batch](https://aws.amazon.com/batch/), and a few other compute resources.",
        "icon": ICON,
        "thumbnail": f"{ABCLI_PUBLIC_PREFIX}/hourglass/workflow.gif?raw=true",
        "url": "https://github.com/kamangir/notebooks-and-scripts/tree/main/blueflow/workflow",
    },
    "scripts": {
        "description": "legacy mechanisms replaced with `@docker eval - <command>` and `@batch eval - <command>`.",
        "icon": ICON,
        "thumbnail": "https://github.com/kamangir/assets/blob/main/nbs/3x4.jpg?raw=true",
        "url": "https://github.com/kamangir/notebooks-and-scripts/tree/main/scripts",
    },
    "template": {
        "description": "",
        "icon": "",
        "thumbnail": "",
        "url": "",
    },
}

items = [
    "{}[`{}`]({}) [![image]({})]({}) {}".format(
        details["icon"],
        feature,
        details["url"],
        details["thumbnail"],
        details["url"],
        details["description"],
    )
    for feature, details in features.items()
    if feature != "template"
]


def build():
    return README.build(
        items=items,
        cols=2,
        path=os.path.join(file.path(__file__), ".."),
        ICON=ICON,
        NAME=NAME,
        VERSION=VERSION,
        MODULE_NAME="blueflow",
        REPO_NAME=REPO_NAME,
    ) and README.build(
        items=workflow_items,
        cols=4,
        path=os.path.join(file.path(__file__), "workflow"),
        NAME=NAME,
        VERSION=VERSION,
        REPO_NAME=REPO_NAME,
    )



from typing import Dict

from blueflow.workflow.runners import RunnerType
from blueflow.workflow.runners.aws_batch import AWSBatchRunner
from blueflow.workflow.runners.local import LocalRunner
from blueflow.workflow.runners.generic import GenericRunner

runner_class: Dict[RunnerType, GenericRunner] = {
    RunnerType.AWS_BATCH: AWSBatchRunner,
    RunnerType.GENERIC: GenericRunner,
    RunnerType.LOCAL: LocalRunner,
}



from sagemaker import get_execution_role

from blueflow.logger import logger

# https://github.com/aws/sagemaker-python-sdk/issues/300#issuecomment-1431539831
role = "arn:aws:iam::120429650996:role/service-role/AmazonSageMaker-ExecutionRole-20231022T170206"

try:
    role = get_execution_role()
except:
    pass

logger.info(f"sagemaker role: {role}")
